/* Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "apr_strings.h"

#include "ap_config.h"
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"

#include <unistd.h>
#include <sys/types.h>
#include <grp.h>
#include <pwd.h>

typedef struct {
    int authoritative;
} authz_group_user_config_rec;

static void *create_authz_group_user_dir_config(apr_pool_t *p, char *d)
{
    authz_group_user_config_rec *conf = apr_palloc(p, sizeof(*conf));

    conf->authoritative = 1; /* keep the fortress secure by default */
    return conf;
}

static const command_rec authz_group_user_cmds[] =
{
    AP_INIT_FLAG("AuthzGroupUserAuthoritative", ap_set_flag_slot,
                 (void *)APR_OFFSETOF(authz_group_user_config_rec, authoritative),
                 OR_AUTHCFG,
                 "Set to 'Off' to allow access control to be passed along to "
                 "lower modules if the 'require user', 'require valid-user' "
                 "'require group' or 'require valid-group' "
                 "statement is not met. (default: On)."),
    {NULL}
};

module AP_MODULE_DECLARE_DATA authz_group_user_module;

static apr_status_t check_user_in_group(request_rec *r, char *user, const char* group) {
    struct group grp,*grp_t;
    struct passwd pwd,*pwd_t;
    int GRP_LEN=sysconf(_SC_GETGR_R_SIZE_MAX);
    int PWD_LEN=sysconf(_SC_GETPW_R_SIZE_MAX);
    char grp_buffer[GRP_LEN];
    char pwd_buffer[PWD_LEN];
    int grp_e,pwd_e;

#if _POSIX_PTHREAD_SEMANTICS
    pwd_e = getpwnam_r(user, &pwd, pwd_buffer, PWD_LEN, &pwd_t);
    if (pwd_e) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
            "getpwnam_r(%s,...) failed with error '%s'",
            user, strerror(pwd_e));
        return grp_e;
    }
#else
    pwd_t = getpwnam_r(user, &pwd, pwd_buffer, PWD_LEN);
#endif
    if (!pwd_t) { 
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                        "User %s does not exist locally", user);
        return APR_ENOENT;
    }
#if _POSIX_PTHREAD_SEMANTICS
    grp_e = getgrgid_r(pwd.pw_gid, &grp, grp_buffer, GRP_LEN, &grp_t);
    if (grp_e) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
            "getgrgid_r(...) failed with error '%s'",
            strerror(grp_e));
        return grp_e;
    }
#else
    grp_t = getgrgid_r(pwd->pw_gid, &grp, grp_buffer, GRP_LEN);
#endif
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                        "User %s has primary group id/name %d/%s, required group is %s",
                        user,pwd_t?pwd.pw_gid:-1,grp.gr_name?grp.gr_name:"NULL",group?group:"NULL");
    if (grp_t) {
        /* primary group */
        if (!strcmp(group,grp.gr_name)) 
            return APR_SUCCESS;
    } 
    /* secondary groups */ 
#if _POSIX_PTHREAD_SEMANTICS
    grp_e = getgrnam_r(group, &grp, grp_buffer, GRP_LEN, &grp_t);
    if (grp_e) {
        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
            "getgrnam_r(...) failed with error '%s'",
            strerror(grp_e));
        return grp_e;
    }
#else
    grp_t = getgrnam_r(group, &grp, grp_buffer, GRP_LEN);
#endif
    if (grp_t) {
        /* members of required group */
        int i=0;
        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                            "Check members of group %s",group); 
        while (grp.gr_mem[i]) { 
            ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                            " %s is member of group %s",grp.gr_mem[i],group); 
            if (!strcmp(group,grp.gr_mem[i])) {
                return APR_SUCCESS;
            }
            i++;
        }
    } 
    return APR_ENOENT;
}

static int check_group_user_access(request_rec *r)
{
    authz_group_user_config_rec *conf = ap_get_module_config(r->per_dir_config,
                                                       &authz_group_user_module);
    char *user = r->user;
    int m = r->method_number;
    int required_group_user = 0;
    register int x;
    const char *t, *w;
    const apr_array_header_t *reqs_arr = ap_requires(r);
    require_line *reqs;

    /* BUG FIX: tadc, 11-Nov-1995.  If there is no "requires" directive,
     * then any user will do.
     */
    if (!reqs_arr) {
        return DECLINED;
    }
    reqs = (require_line *)reqs_arr->elts;

    for (x = 0; x < reqs_arr->nelts; x++) {

        if (!(reqs[x].method_mask & (AP_METHOD_BIT << m))) {
            continue;
        }

        t = reqs[x].requirement;
        w = ap_getword_white(r->pool, &t);
        if (!strcasecmp(w, "valid-user")) {
            errno = 0;
            if (getpwnam(user)) {
                ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                        "User %s is a valid local user", user);
                return OK;
            }
            if (errno) 
                ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                      "access to %s failed, reason: getpwnam(%s) failed with error '%s'", 
                      r->uri, user,strerror(errno));
            return DECLINED;
        }
        if (!strcasecmp(w, "user")) {
            /* And note that there are applicable requirements
             * which we consider ourselves the owner of.
             */
            required_group_user = 1;
            while (t[0]) {
                w = ap_getword_conf(r->pool, &t);
                if (!getpwnam(w)) {
                    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                        "User %s is not a valid local user", w);
                    continue ;
                }
                if (!strcmp(user, w) && getpwnam(user)) {
                    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                                   "User %s is a local user", user);
                    return OK;
                }
            }
        }

        if (!strcasecmp(w, "group")) {
            /* And note that there are applicable requirements
             * which we consider ourselves the owner of.
             */
            required_group_user = 1;
            while (t[0]) {
                w = ap_getword_conf(r->pool, &t);
                if (check_user_in_group(r, user, w) == APR_SUCCESS) {
                    return OK;
                }
            }
        }

    }

    if (!required_group_user) {
        /* no applicable requirements */
        return DECLINED;
    }

    if (!conf->authoritative) {
        return DECLINED;
    }

    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
                  "access to %s failed, reason: user '%s' does not meet "
                  "'require'ments for user/valid-user to be allowed access",
                  r->uri, user);

    ap_note_auth_failure(r);
    return HTTP_UNAUTHORIZED;
}

static void register_hooks(apr_pool_t *p)
{
    ap_hook_auth_checker(check_group_user_access, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA authz_group_user_module =
{
    STANDARD20_MODULE_STUFF,
    create_authz_group_user_dir_config, /* dir config creater */
    NULL,                         /* dir merger --- default is to override */
    NULL,                         /* server config */
    NULL,                         /* merge server config */
    authz_group_user_cmds,        /* command apr_table_t */
    register_hooks                /* register hooks */
};
